// Copyright (C) 2006 Michael Van Canneyt und Toolbox Verlag, Alle Rechte Vorbehalten
unit sqldbsession;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, websession, SqlDB, DB, httpdefs;
  
Type

  { TSQLDBSession }

  TSQLDBSession = Class(TCustomSession)
  Private
    FConnection: TSQLConnection;
    FTransaction : TSQLTransaction;
    FQuery : TSQLQuery;
    FSessionStarted : Boolean;
    FSessionCookie: String;
    FTerminated :Boolean;
    FVariables : TStrings;
    SID : String;
    procedure SetConnection(const AValue: TSQLConnection);
  Protected
    Procedure CheckSession;
    Function GetSessionID : String; override;
    Function GetSessionVariable(VarName : String) : String; override;
    procedure SetSessionVariable(VarName : String; const AValue: String); override;
    // SQLDB methods.
    Function  GetSessionData(Const ASID : String; Var LastSeen : TDateTime; Var ATimeOut : Integer) : Boolean;
    Function  DeleteSession(ASID : String) : Boolean;
    Procedure CreateSession(ASID : String);
    Procedure UpdateSessionTime(ASID : String);
    Procedure DeleteVariable(Const ASID,VariableName : String);
    Procedure InsertVariable(Const ASID,AVArName,AVAlue : String);
    Procedure UpdateVariable(Const ASID,AVArName,AVAlue : String);
    Procedure LoadSessionData(Const ASID : String;List : TStrings);

  Public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Procedure Terminate; override;
    Procedure Notification (AComponent : TComponent; Operation : TOperation); override;
    Procedure UpdateResponse(AResponse : TResponse); override;
    
    Procedure InitSession(ARequest : TRequest; OnNewSession, OnExpired: TNotifyEvent); override;
    Procedure InitResponse(AResponse : TResponse); override;
  Published
    Property Connection : TSQLConnection Read FConnection Write SetConnection;
    Property SessionCookie : String Read FSessionCookie Write FSessionCookie;
  end;

  ESession = Class(HTTPError);

implementation

{$ define cgidebug}
{$ifdef cgidebug}
uses dbugintf;
{$endif cgidebug}


Resourcestring
  SErrNoConnection = '%s: No connection set';

Const
  SFPWebSession = 'FPWebSession';

{ TSQLDBSession }

procedure TSQLDBSession.SetConnection(const AValue: TSQLConnection);
begin
  if FConnection=AValue then
    exit;
  If Assigned(FConnection) then
    FConnection.RemoveFreeNotification(Self);
  FConnection:=AValue;
  If Assigned(FConnection) then
    FConnection.FreeNotification(Self);
end;

Procedure TSQLDBSession.Notification (AComponent : TComponent; Operation : TOperation);

begin
  Inherited;
  if (Operation=opRemove) and (AComponent=FConnection) then
    begin
    FConnection:=Nil;
    If Assigned(FQuery) then
      FQuery.Database:=Nil;
    If Assigned(FTransaction) then
      FTransaction.Database:=Nil;
    end;
end;


constructor TSQLDBSession.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  If Not (csDesigning in COmponentState) then
    begin
    FQuery:=TSQLQuery.Create(Self);
    FTransaction:=TSQLTransaction.Create(Self);
    FQuery.Transaction:=FTransaction;
    FVariables:=TStringList.Create;
    end;
end;

destructor TSQLDBSession.Destroy;
begin
  FreeAndNil(FVariables);
  FreeAndNil(FQuery);
  FreeAndNil(FTransaction);
end;

procedure TSQLDBSession.CheckSession;
begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.CheckSession');{$endif cgidebug}
  If Not Assigned(FConnection) then
    raise ESession.CreateFmt(SErrNoConnection,[Name]);
  If (SessionCookie='') then
    SessionCookie:=SFPWebSession;
  FConnection.Connected:=True;
  If FQuery.Database<>FConnection then
    begin
{$ifdef cgidebug}SendDebug('Setting database');{$endif cgidebug}
    FQuery.Database:=FConnection;
    FTransaction.Database:=FConnection;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.CheckSession');{$endif cgidebug}
end;

function TSQLDBSession.GetSessionID: String;
begin
  If (SID='') then
    SID:=inherited GetSessionID;
  Result:=SID;
end;

function TSQLDBSession.GetSessionVariable(VarName: String): String;
begin
  CheckSession;
  Result:=FVariables.Values[VarName];
end;

procedure TSQLDBSession.SetSessionVariable(VarName: String; const AValue: String
  );

Var
  I : Integer;

begin
  CheckSession;
  I:=FVariables.IndexOfName(VarName);
  If (I=-1) then
    begin
    if (AValue<>'') then
      InsertVariable(SID,VarName,AValue)
    end
  else
    UpdateVariable(SID,VarName,AValue);
  FVariables.Values[VarName]:=AValue;
end;


procedure TSQLDBSession.Terminate;
begin
  FTerminated:=True;
  FSessionStarted:=False;
  DeleteSession(SID);
  SID:='';
  FVariables.Clear;
end;

procedure TSQLDBSession.UpdateResponse(AResponse: TResponse);
begin
  If Assigned(FTransaction) and FTransaction.Active then
    FTransaction.Commit;
end;



procedure TSQLDBSession.InitSession(ARequest: TRequest; OnNewSession,
  OnExpired: TNotifyEvent);

Var
  L   : TDateTime;
  T   : Integer;
  S : String;

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.InitSession');{$endif cgidebug}
  CheckSession;
  S:=ARequest.CookieFields.Values[SessionCookie];
  // have session cookie ?
  If (S<>'') and GetSessionData(S,L,T) then
    begin
    If ((Now-L)>(T/(24*60))) then
      begin
      // Expire session.
      If Assigned(OnExpired) then
        OnExpired(Self);
      DeleteSession(S);
      S:='';
      end
    else
      SID:=S;
    end;
  If (S='') then
    begin
    If Assigned(OnNewSession) then
      OnNewSession(Self);
    GetSessionID;
    CreateSession(SID);
    FVariables.Clear;
    end
  else
    begin
    UpdateSessionTime(SID);
    LoadSessionData(SID,FVariables);
    end;
  FSessionStarted:=True;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.InitSession');{$endif cgidebug}
end;

procedure TSQLDBSession.InitResponse(AResponse: TResponse);
Var
  C : TCookie;
begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.InitResponse');{$endif}
  If FSessionStarted then
    begin
    C:=AResponse.Cookies.Add;
    C.Name:=SessionCookie;
    C.Value:=SID;
     end
  else If FTerminated then
    begin
    C:=AResponse.Cookies.Add;
    C.Name:=SessionCookie;
    C.Value:='';
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.InitResponse');{$endif}
end;

{ ---------------------------------------------------------------------
  SQLDB specific methods
  ---------------------------------------------------------------------}

Const
  SGetSessionData        = 'SELECT SE_LASTSEEN,SE_TIMEOUT FROM SESSIONS WHERE (SE_SESSION=''%s'');';
  SCreateSessionData     = 'INSERT INTO SESSIONS (SE_SESSION,SE_TIMEOUT) VALUES (''%s'',%d);';
  SDeleteSessionData     = 'DELETE FROM SESSIONS WHERE (SE_SESSION=''%s'')';
  SDeleteSessionVariable = 'DELETE FROM SESSIONDATA WHERE (SD_NAME=''%s'') AND (SE_SESSION_FK=(SELECT(SE_ID) FROM SESSIONS WHERE (SE_SESSION=''%s'')))';
  SUpdateSessionTime     = 'UPDATE SESSIONS SET SE_LASTSEEN=''NOW'' WHERE (SE_SESSION=''%s'');';
  SInsertVariable        = 'INSERT INTO SESSIONDATA (SD_SESSION_FK,SD_NAME,SD_VALUE) VALUES '+
                           '((SELECT SE_ID FROM SESSIONS WHERE (SE_SESSION=''%s'')),''%s'',''%s'');';
  SUpdateVariable        = 'UPDATE SESSIONDATA SET SD_VALUE=''%s'' WHERE '+
                           ' (SD_SESSION_FK=(SELECT SE_ID FROM SESSIONS WHERE (SE_SESSION=''%s'')))'+
                           ' AND (SD_NAME=''%s'');';
  SGetVariables          = 'SELECT SD_NAME,SD_VALUE FROM SESSIONS '+
                           'LEFT JOIN SESSIONDATA ON (SD_SESSION_FK=SE_ID)'+
                           ' WHERE (SE_SESSION=''%s'')';

Procedure TSQLDBSession.DeleteVariable(Const ASID,VariableName : String);

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.DeleteVariable');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SDeleteSessionVariable,[ASID,VariableName]);
    ExecSQL;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.DeleteVariable');{$endif}
end;

Function TSQLDBSession.GetSessionData(Const ASID : String; Var Lastseen : TDateTime; Var ATimeOut : Integer) : Boolean;


begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.GetSessionData');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SGetSessionData,[ASID]);
    Open;
    Try
      Result:=Not (EOF AND BOF);
      if Result then
        begin
        LastSeen:=FieldByName('SE_LASTSEEN').AsDateTime;
        ATimeOut:=FieldByName('SE_TIMEOUT').AsInteger;
        end;
    Finally
      Close;
    end;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.GetSessionData');{$endif}
end;

Function TSQLDBSession.DeleteSession(ASID : String) : Boolean;

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.DeleteSession');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SDeleteSessionData,[ASID]);
    ExecSQL;
    Result:=True;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.DeleteSession');{$endif}
end;

Procedure TSQLDBSession.CreateSession(ASID : String);

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.CreateSession');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SCreateSessionData,[ASID,TimeOutMinutes]);
{$ifdef cgidebug}SendDebug('Executing query:'+SQL.Text);{$endif}
    ExecSQL;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.CreateSession');{$endif}
end;

Procedure TSQLDBSession.UpdateSessionTime(ASID : String);

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.UpdateSessionTime');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SUpdateSessionTime,[ASID]);
{$ifdef cgidebug}SendDebug('Executing query:'+SQL.Text);{$endif}
    ExecSQL;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.UpdateSessionTime');{$endif}
end;

Function Escape(S : String): String;

begin
  Result:=StringReplace(S,'''','''''',[rfReplaceAll]);
end;

Procedure TSQLDBSession.InsertVariable(Const ASID,AVArName,AVAlue : String);

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.InsertVariable');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SInsertVariable,[ASID,Escape(AVarName),Escape(AValue)]);
{$ifdef cgidebug}SendDebug('Executing query:'+SQL.Text);{$endif}

    ExecSQL;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.InsertVariable');{$endif}
end;

Procedure TSQLDBSession.UpdateVariable(Const ASID,AVArName,AVAlue : String);

begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.UpdateVariable');{$endif}
  With FQuery do
    begin
    SQL.Text:=Format(SUpdateVariable,[Escape(AValue),ASID,Escape(AVarName)]);
{$ifdef cgidebug}SendDebug('Executing query:'+SQL.Text);{$endif}
    ExecSQL;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.UpdateVariable');{$endif}
end;

Procedure TSQLDBSession.LoadSessionData(Const ASID : String;List : TStrings);

Var
  FN,FV : TField;


begin
{$ifdef cgidebug}SendMethodEnter('TSQLDBSession.LoadSessionData');{$endif}
{$ifdef cgidebug}SendDebug('Clearing list');{$endif}
  List.Clear;
  With FQuery do
    begin
{$ifdef cgidebug}SendDebug('Setting SQL');{$endif}
    SQL.Text:=Format(SGetVariables,[ASID]);
{$ifdef cgidebug}SendDebug('SQL : '+SQL.Text);{$endif}
    Open;
    Try
{$ifdef cgidebug}SendDebug('Opened query');{$endif}
      FN:=Fields.FieldByName('SD_NAME');
      FV:=Fields.FieldByName('SD_VALUE');
{$ifdef cgidebug}SendDebug('Looping data and adding variables');{$endif}
      While not EOF do
        begin
{$ifdef cgidebug}SendDebug(FN.AsString+'='+FV.AsString);{$endif}
        List.Add(FN.AsString+'='+FV.AsString);
        Next;
        end;
    Finally
      Close;
    end;
    end;
{$ifdef cgidebug}SendMethodExit('TSQLDBSession.LoadSessionData');{$endif}
end;


end.

